@@ -42,6 +42,7 @@
   #:use-module (dzn vm util)
   #:export (%exploring?
             did-provides-out?
+            extend-trace
             filter-compliance-error
             filter-error
             filter-illegal+implicit-illegal
@@ -50,11 +51,14 @@
             filter-match-error
             flush-async
             flush-async-trace
+            flush-defer
             interactive?
             livelock?
             mark-livelock-error
             run-async
             run-async-event
+            run-defer-event
+            run-flush
             run-external
             run-external-q
             run-external-modeling
@@ -695,7 +699,8 @@ until RTC?."
                                                               trigger))
                (instance (.container component-port))
                (collateral-blocked? (or (blocked-on-boundary? pc)
-                                        (and (get-handling pc instance)
+                                        (and instance
+                                             (get-handling pc instance)
                                              (blocked-port pc instance))))
                (traces (if collateral-blocked? traces
                            (append-map (cute run-flush <> instance) traces))))
@@ -756,6 +761,64 @@ until RTC?."
   (let ((pc (clone pc #:trail (cons event (.trail pc)))))
     (flush-async pc)))
 
+(define-method (run-defer-event (pc <program-counter>) event)
+  (%debug "run-defer-event ~a pc: ~s\n" event pc)
+  (let* ((pc (prune-defer pc))
+         (defer (.defer pc))
+         (trail (.trail pc))
+         (trail (if (equal? event "<defer>") trail
+                    (cons event trail))))
+    (define (livelock? trace)
+      (let* ((new-pc (car trace))
+             (pc (clone pc #:trail trail)))
+        (and (requires-trigger? event)
+             (rtc-program-counter-equal? new-pc pc)
+             (>= (length (.defer new-pc)) (length  defer))
+             (1- (length trace)))))
+    (if (null? defer) '()
+        (let* ((defer-pc (car defer))
+               (statement (.statement defer-pc))
+               (statement (.statement statement))
+               (instance (.instance defer-pc))
+               (defer-qout (make <defer-qout>
+                             #:statement statement
+                             #:location (.location statement)))
+               (defer-qout (clone defer-qout #:parent (.parent statement)))
+               (pc (clone pc
+                          #:defer (cdr defer)
+                          #:running-defer? instance
+                          #:instance instance
+                          #:statement defer-qout
+                          #:trigger (.trigger defer-pc)
+                          #:trail trail))
+               (pc (graft-locals pc defer-pc))
+               (traces (run-to-completion pc 'rtc))
+               (blocked traces
+                        (partition (compose blocked-on-boundary? car) traces))
+               (traces (map (cute rewrite-trace-head
+                                  (cute clone <> #:running-defer? #f) <>)
+                            traces))
+               (traces (append blocked traces))
+               (livelock traces (partition livelock? traces))
+               (livelock (map (cute mark-livelock-error <> <>)
+                              livelock
+                              (map livelock? livelock))))
+          (append livelock traces)))))
+
+(define-method (flush-defer (pc <program-counter>))
+  (let* ((event "<defer>")
+         (traces (run-defer-event pc event))
+         (blocked? (pair? (.blocked pc)))
+         (traces (if (not blocked?) traces
+                     (append-map
+                      (cute run-to-completion*-context-switch <> event)
+                      traces)))
+         (traces flush (partition (compose null? .defer car) traces)))
+    (append traces (append-map flush-defer flush))))
+
+(define-method (flush-defer trace)
+  (extend-trace trace flush-defer))
+
 (define-method (run-external-q (pc <program-counter>) (instance <runtime:port>))
   (let* ((pc trigger (dequeue-external pc instance))
          (q-out (make <q-out> #:trigger trigger))
@@ -802,17 +865,26 @@ until RTC?."
     (let ((async-traces (if (null? (.async pc)) '()
                             (flush-async-event pc event))))
       (if (pair? async-traces) async-traces
-          (let* ((pcs (cons pc (run-external-modeling pc event)))
-                 (port (.port (string->trigger event)))
-                 (blocked-port (and=> (blocked-on-boundary? pc event) .ast))
-                 (traces (append-map (cute run-requires <> event) pcs)))
-            (if (or (not port) (not (ast:eq? port blocked-port))) traces
-                (filter (negate modeling?) traces))))))
+          (let* ((defer? (and (pair? (.defer pc))
+                              (not (%strict?))))
+                 (defer-traces (if (not defer?) '()
+                                   (run-defer-event pc event))))
+            (append
+             defer-traces
+             (let* ((pcs (cons pc (run-external-modeling pc event)))
+                    (port (.port (string->trigger event)))
+                    (blocked-port (and=> (blocked-on-boundary? pc event)
+                                         .ast))
+                    (traces (append-map (cute run-requires <> event) pcs)))
+               (if (or (not port) (not (ast:eq? port blocked-port))) traces
+                   (filter (negate modeling?) traces))))))))
    ((provides-trigger? event)
     (if (blocked-on-boundary-provides? pc event) (illegal-trace pc)
         (run-to-completion pc event)))
    ((async-event? pc event)
     (flush-async-event pc event))
+   ((defer-event? pc event)
+    (run-defer-event pc event))
    ((and (eq? event #f) (pair? (.async pc)))
     (flush-async pc))
    (else
@@ -821,14 +893,36 @@ until RTC?."
 (define-method (run-to-completion* trace event)
   (extend-trace trace (cute run-to-completion* <> event)))
 
+(define-method (defer-run-to-completion-rtc (pc <program-counter>))
+  (let* ((instance (.running-defer? pc))
+         (pc (clone pc #:running-defer? #f))
+         (instance (if (is-a? instance <runtime:component>) instance
+                       (.container (runtime:other-port instance))))
+         (traces (run-to-completion pc 'rtc))
+         (traces (map (cute rewrite-trace-head
+                            (cute clone <> #:instance instance) <>)
+                      traces))
+         (done traces
+               (partition
+                (disjoin
+                 (compose blocked-on-boundary? car)
+                 (compose (conjoin (cute get-handling <> instance)
+                                   (cute blocked-port <> instance))
+                          car)
+                 (compose (is-status? <error>) car))
+                traces)))
+    (append done
+            (append-map run-flush traces))))
+
 (define-method (run-to-completion*-context-switch (pc <program-counter>) event)
-  (%debug "run-to-completion*-switch-context: ~a\n" event)
+  (%debug "run-to-completion*-context-switch: ~a\n" event)
   (let* ((orig-pc pc)
          (blocked-on-action? (blocked-on-action? pc event))
          (pc (if (and (blocked-on-boundary? pc event)
                       (or blocked-on-action? (return-trigger? event)))
                  (blocked-on-boundary-switch-context pc event)
                  pc))
+         (bob-switched? (not (eq? pc orig-pc)))
          (switched? (not (eq? pc orig-pc)))
          (trigger? (or (is-a? (%sut) <runtime:port>)
                        (provides-trigger? event)
@@ -846,7 +940,14 @@ until RTC?."
                         (provides-trigger? event)))
          (pc (if (or trigger? (not switched?) (rtc-event? event)) pc
                  (clone pc #:trail (cons event (.trail pc)))))
-         (traces (if skip-rtc? (run-to-completion* pc event)
-                     (run-to-completion pc 'rtc))))
+         (traces (cond ((.running-defer? pc)
+                        (defer-run-to-completion-rtc pc))
+                       (skip-rtc?
+                        (run-to-completion* pc event))
+                       (else
+                        (run-to-completion pc 'rtc)))))
     (if (or skip-rtc? (not trigger?)) traces
         (append-map (cute run-to-completion* <> event) traces))))
+
+(define-method (run-to-completion*-context-switch trace event)
+  (extend-trace trace (cute run-to-completion*-context-switch <> event)))
