@@ -79,6 +79,7 @@
             is-status?
             label?
             labels
+            make-implicit-illegal
             make-pc
             make-system-state
             modeling-names
@@ -357,17 +358,27 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
     (values input pc)))
 
 (define-method (labels)
-  (if (is-a? (%sut) <runtime:port>)
-      (let* ((interface ((compose .type .ast %sut)))
-             (modeling-names (modeling-names interface)))
-        (append (map .name (ast:in-event* interface))
-                modeling-names))
+  (cond
+   ((is-a? (%sut) <runtime:port>)
+    (let* ((interface ((compose .type .ast %sut)))
+           (modeling-names (modeling-names interface)))
+      (append (map .name (ast:in-event* interface))
+              modeling-names)))
+   (else
+    (let ((ports (filter
+                  (conjoin runtime:boundary-port?
+                           runtime:other-port
+                           (negate (compose (is? <runtime:foreign>)
+                                            .container
+                                            runtime:other-port)))
+                  (%instances))))
       (append-map
        (lambda (p)
-         (map (compose (cute string-append (runtime:dotted-name p) "." <>) .name)
-              (filter (if (or (ast:provides? (.ast p))) ast:in? ast:out?)
+         (map (compose (cute string-append (runtime:dotted-name p) "." <>)
+                       .name)
+              (filter (if (ast:provides? (.ast p)) ast:in? ast:out?)
                       (ast:event* (.ast p)))))
-       (filter runtime:boundary-port? (%instances)))))
+       ports)))))
 
 (define-method (labels (pc <program-counter>))
   (append (labels) (rtc-labels pc) (return-labels pc) (defer-labels pc)))
@@ -497,7 +508,7 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
 (define-method (trigger->system-trigger (o <runtime:port>) trigger)
   (trigger->component-trigger o trigger))
 
-(define-method (trigger->system-trigger (component <runtime:component>) trigger)
+(define-method (trigger->system-trigger (component <runtime:component-model>) trigger)
   (let* ((port-name (.port.name trigger))
          (ports (runtime:runtime-port* component))
          (port (find (compose (cute equal? <> port-name) .name .ast) ports))
@@ -1504,3 +1515,7 @@ See <https://www.gnu.org/licenses/agpl.html>, for more details.
          (pc (if (null? errors) pc
                  (clone pc #:status (car errors)))))
   pc))
+
+(define-method (make-implicit-illegal (pc <program-counter>) (o <ast>))
+  (let ((illegal (make <implicit-illegal-error> #:ast o #:message "illegal")))
+    (clone pc #:previous #f #:status illegal)))
